//
//  MyScene.m
//  SpriteKitSimpleGame
//
//  Created by Main Account on 9/4/13.
//  Copyright (c) 2013 Razeware LLC. All rights reserved.
//

#import "MyScene.h"
#import "GameOverScene.h"

#if TARGET_OS_IPHONE
@import CoreMotion;
#endif

static const uint32_t projectileCategory     =  0x1 << 0;
static const uint32_t monsterCategory        =  0x1 << 1;


@interface MyScene () <SKPhysicsContactDelegate>
@property (nonatomic) SKSpriteNode *player;
@property (nonatomic) NSTimeInterval lastSpawnTimeInterval;
@property (nonatomic) NSTimeInterval lastUpdateTimeInterval;
@property (nonatomic) int monstersDestroyed;
@end

static inline CGPoint rwAdd(CGPoint a, CGPoint b) {
  return CGPointMake(a.x + b.x, a.y + b.y);
}

static inline CGPoint rwSub(CGPoint a, CGPoint b) {
  return CGPointMake(a.x - b.x, a.y - b.y);
}

static inline CGPoint rwMult(CGPoint a, float b) {
  return CGPointMake(a.x * b, a.y * b);
}

static inline float rwLength(CGPoint a) {
  return sqrtf(a.x * a.x + a.y * a.y);
}

// Makes a vector have a length of 1
static inline CGPoint rwNormalize(CGPoint a) {
  float length = rwLength(a);
  return CGPointMake(a.x / length, a.y / length);
}

@implementation MyScene {
#if TARGET_OS_IPHONE
  CMMotionManager *_motionManager;
#endif
  SKAction *_playPewPew;
  SKTextureAtlas *_sprites;
}

-(id)initWithSize:(CGSize)size {
  if (self = [super initWithSize:size]) {
    
#if TARGET_OS_IPHONE
    _sprites = [SKTextureAtlas atlasNamed:@"sprites.atlas"];
#else
    _sprites = [SKTextureAtlas atlasNamed:@"spritesMac.atlas"];
#endif
    
    _playPewPew = [SKAction playSoundFileNamed:@"pew-pew-lei.caf" waitForCompletion:NO];
    
    SKTexture *bgRepeat = [SKTexture textureWithImageNamed:@"bg_repeatable"];
    CGFloat startPoint = 0.0;
    
    while (startPoint < self.size.width) {
      SKSpriteNode *bgImage = [SKSpriteNode spriteNodeWithTexture:bgRepeat];
      bgImage.anchorPoint = CGPointMake(0.0, 0.0);
      bgImage.size = CGSizeMake((self.size.height / bgRepeat.size.height) * bgRepeat.size.width,self.size.height);
      bgImage.position = CGPointMake(startPoint, 0.0);
      [self addChild:bgImage];
      startPoint = startPoint + bgImage.size.width;
    }
    
    self.player = [SKSpriteNode spriteNodeWithImageNamed:@"player"];
    self.player.position = CGPointMake(self.player.size.width/2, self.frame.size.height/2);
    [self addChild:self.player];
    
    self.physicsWorld.gravity = CGVectorMake(0,0);
    self.physicsWorld.contactDelegate = self;
    
#if TARGET_OS_IPHONE
    _motionManager = [[CMMotionManager alloc] init];
    _motionManager.accelerometerUpdateInterval = 0.05;
    [_motionManager startAccelerometerUpdates];
#endif
    
  }
  return self;
}

- (void)addMonster {
  
  SKSpriteNode *monster = [SKSpriteNode spriteNodeWithImageNamed:@"monster"];
  
  monster.physicsBody = [SKPhysicsBody bodyWithRectangleOfSize:monster.size];
  monster.physicsBody.dynamic = YES;
  monster.physicsBody.categoryBitMask = monsterCategory;
  monster.physicsBody.contactTestBitMask = projectileCategory;
  monster.physicsBody.collisionBitMask = 0;
  
  int minY = monster.size.height / 2;
  int maxY = self.frame.size.height - monster.size.height / 2;
  int rangeY = maxY - minY;
  int actualY = (arc4random() % rangeY) + minY;
  
  monster.position = CGPointMake(self.frame.size.width + monster.size.width/2, actualY);
  [self addChild:monster];
  
  int minDuration = 2.0;
  int maxDuration = 4.0;
  int rangeDuration = maxDuration - minDuration;
  int actualDuration = (arc4random() % rangeDuration) + minDuration;
  
  SKAction *actionMove = [SKAction moveTo:CGPointMake(-monster.size.width/2, actualY) duration:actualDuration];
  SKAction *actionMoveDone = [SKAction removeFromParent];
  SKAction *loseAction = [SKAction runBlock:^{
    SKTransition *reveal = [SKTransition flipHorizontalWithDuration:0.5];
    SKScene *gameOverScene = [[GameOverScene alloc] initWithSize:self.size won:NO];
    [self.view presentScene:gameOverScene transition: reveal];
  }];
  [monster runAction:[SKAction sequence:@[actionMove, loseAction, actionMoveDone]]];
  
}

- (void)updateWithTimeSinceLastUpdate:(CFTimeInterval)timeSinceLast {
  
  self.lastSpawnTimeInterval += timeSinceLast;
  if (self.lastSpawnTimeInterval > 1) {
    self.lastSpawnTimeInterval = 0;
    [self addMonster];
  }
}

- (void)update:(NSTimeInterval)currentTime {
  
  CFTimeInterval timeSinceLast = currentTime - self.lastUpdateTimeInterval;
  self.lastUpdateTimeInterval = currentTime;
  if (timeSinceLast > 1) {
    timeSinceLast = 1.0 / 60.0;
    self.lastUpdateTimeInterval = currentTime;
  }
  
  [self updateWithTimeSinceLastUpdate:timeSinceLast];
#if TARGET_OS_IPHONE
  [self updatePlayerWithTimeSinceLastUpdate:timeSinceLast];
#endif
}

#if TARGET_OS_IPHONE
- (void)updatePlayerWithTimeSinceLastUpdate:(CFTimeInterval)timeSinceLast {
  CGFloat kFilteringFactor = 0.75;
  static UIAccelerationValue rollingX = 0, rollingY = 0, rollingZ = 0;
  
  rollingX = (_motionManager.accelerometerData.acceleration.x * kFilteringFactor) +
  (rollingX * (1.0 - kFilteringFactor));
  rollingY = (_motionManager.accelerometerData.acceleration.y * kFilteringFactor) +
  (rollingY * (1.0 - kFilteringFactor));
  rollingZ = (_motionManager.accelerometerData.acceleration.z * kFilteringFactor) +
  (rollingZ * (1.0 - kFilteringFactor));
  
  CGFloat accelX = rollingX;
  
  CGFloat kRestAccelX = 0.6;
  CGFloat kPlayerMaxPointsPerSec = self.size.height*0.5;
  CGFloat kMaxDiffX = 0.2;
  
  CGFloat accelDiffX = kRestAccelX - ABS(accelX);
  CGFloat accelFractionX = accelDiffX / kMaxDiffX;
  CGFloat pointsPerSecX = kPlayerMaxPointsPerSec * accelFractionX;
  
  CGFloat playerPointsPerSecY = pointsPerSecX;
  CGFloat maxY = self.size.height - _player.size.height/2;
  CGFloat minY = _player.size.height/2;
  
  CGFloat newY = _player.position.y + (playerPointsPerSecY * timeSinceLast);
  newY = MIN(MAX(newY, minY), maxY);
  _player.position = CGPointMake(_player.position.x, newY);
}
#else
-(void)keyDown:(NSEvent *)theEvent {
  
  NSString *keyPressed = [theEvent charactersIgnoringModifiers];
  if ([keyPressed length] == 1) {
    //NSLog(@"Key: %c",[keyPressed characterAtIndex:0]);
    unichar charPressed = [keyPressed characterAtIndex:0];
    switch (charPressed) {
      case NSUpArrowFunctionKey:
        [_player runAction:[SKAction moveByX:0.0f y:50.0f duration:0.3]];
        break;
      case NSDownArrowFunctionKey:
        [_player runAction:[SKAction moveByX:0.0f y:-50.0f duration:0.3]];
        break;
      default:
        break;
    }
  }
  
}
#endif

-(void)screenInteractionEndedAtLocation:(CGPoint)location {
  
  [self runAction:_playPewPew];
  
  SKSpriteNode *projectile = [SKSpriteNode spriteNodeWithImageNamed:@"projectile"];
  projectile.position = self.player.position;
  projectile.physicsBody = [SKPhysicsBody bodyWithCircleOfRadius:projectile.size.width/2];
  projectile.physicsBody.dynamic = YES;
  projectile.physicsBody.categoryBitMask = projectileCategory;
  projectile.physicsBody.contactTestBitMask = monsterCategory;
  projectile.physicsBody.collisionBitMask = 0;
  projectile.physicsBody.usesPreciseCollisionDetection = YES;
  
  CGPoint offset = rwSub(location, projectile.position);
  
  if (offset.x <= 0) return;
  
  [self addChild:projectile];
  
  CGPoint direction = rwNormalize(offset);
  
  CGPoint shootAmount = rwMult(direction, 1000);
  
  CGPoint realDest = rwAdd(shootAmount, projectile.position);
  
  float sizeFactor = (self.size.height / 640.0f) * 2;
  float velocity = (480.0/1.0)*sizeFactor;
  float realMoveDuration = self.size.width / velocity;
  SKAction *actionMove = [SKAction moveTo:realDest duration:realMoveDuration];
  SKAction *actionMoveDone = [SKAction removeFromParent];
  [projectile runAction:[SKAction sequence:@[actionMove, actionMoveDone]]];
  
}

- (void)projectile:(SKSpriteNode *)projectile didCollideWithMonster:(SKSpriteNode *)monster {
  NSLog(@"Hit");
  
  SKEmitterNode *emitter = [NSKeyedUnarchiver unarchiveObjectWithFile:[[NSBundle mainBundle] pathForResource:@"SmallExplosion" ofType:@"sks"]];
  emitter.particleTexture = [SKTexture textureWithImageNamed:@"spark"];
  float scaleFactor = (self.size.height / 640.0f) * 2;
  emitter.xScale = scaleFactor;
  emitter.yScale = scaleFactor;
  emitter.position = monster.position;
  emitter.zPosition = 5;
  [emitter runAction:[SKAction sequence:@[[SKAction waitForDuration:1.0],[SKAction removeFromParent]]]];
  [self addChild:emitter];
  
  [projectile removeFromParent];
  [monster removeFromParent];
  self.monstersDestroyed++;
  if (self.monstersDestroyed > 30) {
    SKTransition *reveal = [SKTransition flipHorizontalWithDuration:0.5];
    SKScene *gameOverScene = [[GameOverScene alloc] initWithSize:self.size won:YES];
    [self.view presentScene:gameOverScene transition: reveal];
  }
}

- (void)didBeginContact:(SKPhysicsContact *)contact
{
  SKPhysicsBody *firstBody, *secondBody;
  
  if (contact.bodyA.categoryBitMask < contact.bodyB.categoryBitMask)
  {
    firstBody = contact.bodyA;
    secondBody = contact.bodyB;
  }
  else
  {
    firstBody = contact.bodyB;
    secondBody = contact.bodyA;
  }
  
  if ((firstBody.categoryBitMask & projectileCategory) != 0 &&
      (secondBody.categoryBitMask & monsterCategory) != 0)
  {
    [self projectile:(SKSpriteNode *) firstBody.node didCollideWithMonster:(SKSpriteNode *) secondBody.node];
  }
}

@end
